#!/usr/bin/perl
BEGIN{ push(@INC, $ENV{IS_SCRIPTS} || "/usr/share/metavelvet/ISScripts"); }
use strict;
use warnings;
use Getopt::Std;
use FASTA::FileIO;
use FASTA::Util;
use BLAST::M9FileIO;
use BLAST::Util;
use Sequence::LociUtil;
use Utils::Calc;
use vars qw($opt_h $opt_i $opt_s $opt_g $opt_O $opt_E $opt_F $opt_C $opt_P $opt_B $opt_e $opt_f $opt_c $opt_p $opt_b $opt_R $opt_L $opt_D $opt_T $opt_I);
&getopts('hi:s:g:O:E:F:C:P:Be:f:c:p:bR:L:D:T:I');

my $usage = <<_EOH_;

## Options ###########################################
## Required:
# -i <path>    Working directory
# -s <file>    IS fasta file
# -g <file>    Reference genome FastA file

## Optional (Output):
# -O <path>    Output prefix (default = <work-dir>/annoIS)

## Optional (BLAST IS vs. contig/scaffold):
# -E <double>  BLAST E-value cutoff (default = 1e-5)
# -F <int>     Minimum IS length of BLAST hits (default = 100bp)
# -C <double>  Minimum cover rate of IS (default = 50%)
# -P <double>  Minimum percent identity (default = 90%)
# -B           Use RBH BLAST hits only (default = Use contig->IS best hit only)

## Optional (BLAST IS-adjacent sequences vs. reference):
# -e <double>  BLAST E-value cutoff (default = 1e-5)
# -f <int>     Minimum IS-adjecent sequence length of BLASt hits (default = 100bp)
# -c <double>  Minimum cover rate of IS-adjacent sequence (default = 90%)
# -p <double>  Minimum percent identity (default = 90%)
# -b           Allow multiple hits per a IS-adjacent sequence (default = not allow)

## Optional (IS annotations):
# -R <int>     Minimum reliable node length (default = 500bp)
# -L <int>     Length of IS flanking sequences (default = 200bp)
# -D <int>     Allowable distance between 5' and 3' IS-adjacent sequences (default = 5000bp)

## Optional (Computation):
# -T <int>     Num. threads (default = 1)
# -I           Turn on flag to use inter-mediate results if available (default = no)


## Others:
# -h print help

_EOH_
    ;

die $usage if $opt_h;

# Get command line options
my $workDir = $opt_i or die $usage;
my $isFasta = $opt_s or die $usage;
my $refFile = $opt_g or die $usage;
my $oPrefix = $workDir . '/' . 'annoIS'; $oPrefix = $opt_O if defined $opt_O;
my $ctgFile = $workDir . '/' . 'contigs.fa';
my $nodFile = $oPrefix . '.' . 'IS-nodes.txt';
my $blsFile = $oPrefix . '.' . 'vs-IS.blast';
my $arnFile = $oPrefix . '.' . 'AR-nodes.txt';
my $annFile = $oPrefix . '.' . 'IS-positions.txt';
## BLAST (IS vs. contig/scaffold)
my $evalCut = 1e-5; $evalCut = $opt_E if defined $opt_E;
my $minHLen = 100;  $minHLen = $opt_F if defined $opt_F;
my $minHCov = 50;   $minHCov = $opt_C if defined $opt_C;
my $minIden = 90;   $minIden = $opt_P if defined $opt_P;
my $flgBest = 0;    $flgBest = 1      if defined $opt_B;
## BLAST (IS vs. contig/scaffold)
my $EvalCut = 1e-5; $evalCut = $opt_e if defined $opt_e;
my $MinHLen = 100;  $minHLen = $opt_f if defined $opt_f;
my $MinHCov = 90;   $MinHCov = $opt_c if defined $opt_c;
my $MinIden = 90;   $minIden = $opt_p if defined $opt_p;
my $MultHit = 0;    $MultHit = 1      if defined $opt_b;
## IS annotation
my $rndLen  = 500;  $rndLen  = $opt_R if defined $opt_R;
my $flkLen  = 200;  $flkLen  = $opt_L if defined $opt_L;
my $isaDist = 5000; $isaDist = $opt_D if defined $opt_D;
## Computation
my $numCPUs = 1;    $numCPUs = $opt_T if defined $opt_T;
my $flgIRes = 0;    $flgIRes = 1      if defined $opt_I;

# Run BLAST vs. IS
unless( $flgIRes and (-e $blsFile) ){
    print STDERR "Running BLAST against IS...\n\t";
    BLAST::Util::runBlast( { programName => 'blastn',
			     formatCode  => 8,
			     queryFile   => $isFasta, 
			     subjectFile => $ctgFile, 
			     outputFile  => $blsFile, 
			     maxEvalue   => $evalCut, 
			     numCPUs     => $numCPUs }, "\t" );
    print STDERR "...done.\n\n";
}

unless( $flgIRes and (-e $arnFile) ){
    # Load ISs
    print STDERR "Loading IS fasta ...\n"
	. "\t" . "filename = " . $isFasta . "\n";
    my $isSeqs   = FASTA::FileIO::parse( $isFasta );
    my $isID2seq = FASTA::Util::indexByShortTitle( $isSeqs );
    print STDERR "\t" . scalar(@$isSeqs) . " ISs were loaded.\n"
	. "...done.\n\n";

    # Load BLAST hits
    print STDERR "Loading and filtering BLAST hits (IS vs contig/scaffold) ...\n"
	. "\t" . "filename = " . $blsFile . "\n";
    my $hits = getBlastHits( BLAST::M9FileIO::parse( $blsFile ), $minHLen, $minHCov, $minIden, $flgBest, $isID2seq );
    print STDERR "...done.\n\n";
    
    # Annotate IS center nodes
    print STDERR "Annotating IS nodes ...\n"
	. "\t" . "filename = " . $nodFile . "\n";
    my $nodeID2info = annotateISNodes( $hits, $isID2seq, $flkLen );
    outputNodeInfo( $nodFile, $nodeID2info );
    print STDERR "...done.\n\n";
    
    # Run annoIS program
    my $cmd = "annoIS $workDir $nodFile $arnFile -reliable_length $rndLen -flanking_length $flkLen";
    print STDERR "Running annoIS program ...\n"
	. "\t CMD: " . $cmd . "\n";
    system( $cmd );
    print STDERR "...done.\n";
}

unless( $flgIRes and (-e $annFile) ){
    # Load contigs
    print STDERR "Loading contig fasta ...\n";
    my $ctgSeqs    = FASTA::FileIO::parse( $ctgFile );
    my $nodeID2seq = getNodeID2seqs( $ctgSeqs );
    print STDERR "\t" . scalar(@$ctgSeqs) . " contigs/scaffolds were loaded.\n"
	. "...done.\n\n";

    # Extract IS-adjacent sequences
    print STDERR "Determinig IS positions ...\n";
    my $annoISs        = [];
    my $nodeID2isaInfo = loadNodeID2isaInfo( $arnFile );
    foreach my $nodeID (sort{$a<=>$b} keys %$nodeID2isaInfo){
	my $isaInfo  = $nodeID2isaInfo->{$nodeID};
	my $isName   = $isaInfo->{isName};
	print STDERR "----------------------------------------------\n"
	    . "\t" . $nodeID . "\t" . $isName . "\n";

	my $isaSeqs5 = getIsaSeqs5( $isaInfo->{arNodes5}, $nodeID2seq, $flkLen );
	my $isaFile5 = $oPrefix . ".$nodeID.isa$flkLen.5.fa";
	FASTA::FileIO::output( $isaFile5, $isaSeqs5 );
	my $isaID2seq5 = FASTA::Util::indexByShortTitle( $isaSeqs5 );
	print STDERR "\t" . scalar(@$isaSeqs5) . " IS-adjacent sequences were extracted (5' direction).\n";

	my $blsFile5 = $oPrefix . ".$nodeID.isa$flkLen.5.blast";
	BLAST::Util::runBlast( { programName => 'blastn',
				 formatCode  => 8,
				 queryFile   => $isaFile5, 
				 subjectFile => $refFile, 
				 outputFile  => $blsFile5, 
				 maxEvalue   => $EvalCut, 
				 numCPUs     => $numCPUs }, "\t\t" );
	my $hits5    = GetBlastHits( BLAST::M9FileIO::parse( $blsFile5 ), $MinHLen, $MinHCov, $MinIden, $MultHit, $isaID2seq5 );
	print STDERR "----------------------------------------------\n";
	
	my $isaSeqs3 = getIsaSeqs3( $isaInfo->{arNodes3}, $nodeID2seq, $flkLen );
	my $isaFile3 = $oPrefix . ".$nodeID.isa$flkLen.3.fa";
	FASTA::FileIO::output( $isaFile3, $isaSeqs3 );
	my $isaID2seq3 = FASTA::Util::indexByShortTitle( $isaSeqs3 );
	print STDERR "\t" . scalar(@$isaSeqs3) . " IS-adjacent sequences were extracted (3' direction).\n";

	my $blsFile3 = $oPrefix . ".$nodeID.isa$flkLen.3.blast";
	BLAST::Util::runBlast( { programName => 'blastn',
				 formatCode  => 8,
				 queryFile   => $isaFile3, 
				 subjectFile => $refFile, 
				 outputFile  => $blsFile3, 
				 maxEvalue   => $EvalCut, 
				 numCPUs     => $numCPUs }, "\t\t" );
	my $hits3    = GetBlastHits( BLAST::M9FileIO::parse( $blsFile3 ), $MinHLen, $MinHCov, $MinIden, $MultHit, $isaID2seq3 );
	print STDERR "----------------------------------------------\n";

	my $_annoISs  = annotateIS( $hits5, $hits3, $isaDist );
	foreach my $annoIS (@$_annoISs){
	    $annoIS->{nodeID} = $nodeID;
	    $annoIS->{isName} = $isName;
	}
	push( @$annoISs, @$_annoISs );

	print STDERR "----------------------------------------------\n\n";
    }
    outputAnnotations( $annFile, $annoISs );
    print STDERR "...done.\n\n";
}


sub getBlastHits {
    my $hits     = shift;
    my $minHLen  = shift;
    my $minHCov  = shift;
    my $minIden  = shift;
    my $flgBest  = shift;
    my $isID2seq = shift;
    print STDERR "\t" . "#. total hits         = " . scalar(@$hits) . "\n";
    $hits        = getLongHits( $hits, $minHLen );
    print STDERR "\t" . "#. long hits          = " . scalar(@$hits) . "\n";
    $hits        = getCoveredHits( $hits, $minHCov, $isID2seq );
    print STDERR "\t" . "#. covered hits       = " . scalar(@$hits) . "\n";
    $hits        = getHighIdentityHits( $hits, $minIden );
    print STDERR "\t" . "#. high identity hits = " . scalar(@$hits) . "\n";
    $hits        = getSubjectBestHits( $hits );
    print STDERR "\t" . "#. subject best hits  = " . scalar(@$hits) . "\n";
    if( $flgBest ){
	$hits    = getQueryBestHits( $hits );
	print STDERR "\t" . "#. query best hits    = " . scalar(@$hits) . "\n";
    }
    return $hits;
}

sub GetBlastHits {
    my $hits     = shift;
    my $minHLen  = shift;
    my $minHCov  = shift;
    my $minIden  = shift;
    my $MultHit  = shift;
    my $isID2seq = shift;
    print STDERR "\t" . "#. total hits         = " . scalar(@$hits) . "\n";
    $hits        = getLongHits( $hits, $minHLen );
    print STDERR "\t" . "#. long hits          = " . scalar(@$hits) . "\n";
    $hits        = getCoveredHits( $hits, $minHCov, $isID2seq );
    print STDERR "\t" . "#. covered hits       = " . scalar(@$hits) . ' ( cutoff = ' . $minHCov . ' %)' . "\n";
    $hits        = getHighIdentityHits( $hits, $minIden );
    print STDERR "\t" . "#. high identity hits = " . scalar(@$hits) . "\n";
    if( !$MultHit ){
	$hits        = getQueryBestHits( $hits );
	print STDERR "\t" . "#. query best hits  = " . scalar(@$hits) . "\n";
    }
    return $hits;
}

sub getLongHits {
    my $hits   = shift;
    my $minLen = shift;
    my $longHits = [];
    foreach my $hit (@$hits){
	push( @$longHits, $hit ) if $hit->getQueryLength() >= $minLen;
    }
    return $longHits;
}

sub getCoveredHits {
    my $hits = shift;
    my $minCov = shift;
    my $isID2seq = shift;
    my $coveredHits = [];
    foreach my $hit (@$hits){
	push( @$coveredHits, $hit ) if calcCoverRate($hit, $isID2seq->{$hit->getQueryID()}->getLength()) >= $minCov;
    }
    return $coveredHits;
}

sub getHighIdentityHits {
    my $hits = shift;
    my $minIden = shift;
    my $idenHits = [];
    foreach my $hit (@$hits){
	push( @$idenHits, $hit ) if $hit->getPercentIdentity() >= $minIden;
    }
    return $idenHits;
}

sub calcCoverRate {
    my $hit = shift;
    my $len = shift;
    return Utils::Calc::round( $hit->getQueryLength()/$len*100, 2 );
}

sub getQueryBestHits {
    my $hits = shift;
    my $isID2hit = {};
    foreach my $hit (@$hits){
	my $isID = $hit->getQueryID();
	if( !defined($isID2hit->{$isID}) ){
	    $isID2hit->{$isID} = $hit;
	} else {
	    $isID2hit->{$isID} = $hit if $hit->getBitScore() > $isID2hit->{$isID}->getBitScore();
	}
    }
    my $bestHits = [];
    foreach my $isID (sort keys %$isID2hit){
	push( @$bestHits, $isID2hit->{$isID} );
    }
    return $bestHits;
}

sub getSubjectBestHits {
    my $hits = shift;
    my $ctgID2hit = {};
    foreach my $hit (@$hits){
	my $ctgID = $hit->getSubjectID();
	if( !defined($ctgID2hit->{$ctgID}) ){
	    $ctgID2hit->{$ctgID} = $hit;
	} else {
	    $ctgID2hit->{$ctgID} = $hit if $hit->getBitScore() > $ctgID2hit->{$ctgID}->getBitScore();
	}
    }
    my $bestHits = [];
    foreach my $ctgID (sort keys %$ctgID2hit){
	push( @$bestHits, $ctgID2hit->{$ctgID} );
    }
    return $bestHits;
}


sub annotateISNodes {
    my $hits        = shift;
    my $isID2seq    = shift;
    my $flkLen      = shift;
    my $nodeID2info = {};
    foreach my $hit (@$hits){
	my $nodeID = getNodeIDFromContigID( $hit->getSubjectID() );
	my $isLen  = $isID2seq->{$hit->getQueryID()}->getLength();
	$nodeID2info->{$nodeID} = getNodeInfo( $hit, $isLen, $flkLen );
    }
    return $nodeID2info;
}

sub getNodeInfo {
    my $hit    = shift;
    my $isLen  = shift;
    my $flKLen = shift;
    return { isName  => $hit->getQueryID(),
	     offset5 => $hit->getSubjectStart(), 
	     offset3 => $hit->getSubjectEnd(),
	     expand5 => getExpand5( $hit, $flkLen ),
	     expand3 => getExpand3( $hit, $isLen, $flkLen ),
	     length  => getNodeLengthFromContigID($hit->getSubjectID()) };
}

sub getExpand5 {
    my $hit    = shift;
    my $flkLen = shift;
    return $hit->getQueryStart() - 1 + $flkLen;
}

sub getExpand3 {
    my $hit    = shift;
    my $isLen  = shift;
    my $flkLen = shift;
    return $isLen - $hit->getQueryEnd() + $flkLen;
}

sub getNodeIDFromContigID {
    my $ctgID = shift;
    return $1 if $ctgID =~ /^NODE_(\d+)_/;
    die "Invalid contig sequence ID was found: $ctgID";
}

sub getNodeLengthFromContigID {
    my $ctgID = shift;
    return $2 if $ctgID =~ /^NODE_(\d+)_length_(\d+)/;
    die "Invalid contig sequence ID was found: $ctgID";
}

sub outputNodeInfo {
    my $file = shift;
    my $nodeID2offsets = shift;
    open OUT, ">$file" or die "Can't open $file to write.";
    print STDERR "\t" . "NodeID" . "\t" . "ISName" . "\t" . "Offset5" . "\t" . "Expand5" . "\t"
	. "Offset3" . "\t" . "Expand3" . "\t" . "NodeLength" . "\n"
	. "\t" . "----" . "\t" . "----" . "\t" . "----" . "\t" . "----" . "\t"
	. "----" . "\t" . "----" . "\t" . "----" . "\n";
    foreach my $nodeID (sort{$a<=>$b} keys %$nodeID2offsets){
	my $offsets = $nodeID2offsets->{$nodeID};
	print OUT $nodeID . "\t" . $offsets->{isName} . "\t" . $offsets->{offset5} . "\t" . $offsets->{expand5} 
	. "\t" . $offsets->{offset3} . "\t" . $offsets->{expand3} . "\t" . $offsets->{length} . "\n";
	print STDERR "\t" . $nodeID . "\t" . $offsets->{isName} . "\t" . $offsets->{offset5} . "\t" . $offsets->{expand5}
	. "\t" . $offsets->{offset3} . "\t" . $offsets->{expand3} . "\t" . $offsets->{length} . "\n";
    }
    close OUT;
}

sub getNodeID2seqs {
    my $ctgSeqs = shift;
    my $nodeID2seq = {};
    foreach my $seq (@$ctgSeqs){
	my $nodeID = getNodeIDFromContigID( $seq->getShortTitle() );
	$nodeID2seq->{$nodeID} = $seq;
    }
    return $nodeID2seq;
}

sub loadNodeID2isaInfo {
    my $file = shift;
    my $nodeID2isaInfo = {};
    open IN, $file or die "Can't open $file to read.";
    while(<IN>){
	chomp;
	my @cols = split("\t", $_);
	if( scalar(@cols) == 4 ){
	    my ($nodeID, $isName, $_arNodes5, $_arNodes3) = @cols;
	    my @arNodes5 = split('\|', $_arNodes5); my $arNodes5 = \@arNodes5;
	    my @arNodes3 = split('\|', $_arNodes3); my $arNodes3 = \@arNodes3;
	    $nodeID2isaInfo->{$nodeID} = { isName => $isName, arNodes5 => $arNodes5, arNodes3 => $arNodes3 };
	}
    }
    close IN;
    return $nodeID2isaInfo;
}

sub getIsaSeqs5 {
    my $arNodes    = shift;
    my $nodeID2seq = shift;
    my $flkLen     = shift;
    my $isaSeqs    = [];
    foreach my $nodeID (@$arNodes){
	my $flagRC = 0;
	if( $nodeID < 0 ){
	    $flagRC = 1;
	    $nodeID *= -1;
	}
	my $seq = $nodeID2seq->{$nodeID};
	my $isaSeq = new FASTA::Sequence();
	$isaSeq->setTitle( $nodeID . "_flanking" . $flkLen . "bp_RC$flagRC" );
	if( $flagRC ){
	    $isaSeq->setSequence( substr($seq->getRCSequence(), $seq->getLength()-$flkLen+1, $flkLen) );
	} else {
	    $isaSeq->setSequence( substr($seq->getSequence(), $seq->getLength()-$flkLen+1, $flkLen) );
	}
	push( @$isaSeqs, $isaSeq );
    }
    return $isaSeqs;
}

sub getIsaSeqs3 {
    my $arNodes    = shift;
    my $nodeID2seq = shift;
    my $flkLen     = shift;
    my $isaSeqs    = [];
    foreach my $nodeID (@$arNodes){
	my $flagRC = 0;
	if( $nodeID < 0 ){
	    $flagRC = 1;
	    $nodeID *= -1;
	}
	my $seq = $nodeID2seq->{$nodeID};
	my $isaSeq = new FASTA::Sequence();
	$isaSeq->setTitle( $nodeID . "_flanking" . $flkLen . "bp_RC$flagRC" );
	if( $flagRC ){
	    $isaSeq->setSequence( substr($seq->getRCSequence(), 0, $flkLen) );
	} else {
	    $isaSeq->setSequence( substr($seq->getSequence(), 0, $flkLen) );
	}
	push( @$isaSeqs, $isaSeq );
    }
    return $isaSeqs;
}

sub annotateIS {
    my $hits5   = shift;
    my $hits3   = shift;
    my $isaDist = shift;
    my $annoISs = [];
    
    my $flag5 = {};
    foreach my $hit5 (@$hits5){
	$flag5->{$hit5->getQueryID()} = 0;
    }
    my $flag3 = {};
    foreach my $hit3 (@$hits3){
	$flag3->{$hit3->getQueryID()} = 0;
    }

    foreach my $hit5 (@$hits5){
	my $isaLocus5 = $hit5->getSubjectLocus();
	foreach my $hit3 (@$hits3){
	    my $isaLocus3 = $hit3->getSubjectLocus();
	    if( my $annoIS = getAnnoISPair( $isaLocus5, $isaLocus3, $isaDist) ){
		$flag5->{$hit5->getQueryID()} = 1;
		$flag3->{$hit3->getQueryID()} = 1;
		push( @$annoISs, $annoIS );
	    }
	}
    }
    print STDERR "\t" . scalar(@$annoISs) . " IS positions were supported by paired IS-adjacent sequences.\n" ;

    foreach my $hit5 (@$hits5){
	if( $flag5->{$hit5->getQueryID()} == 0 ){
	    my $annoIS = getAnnoIS5( $hit5->getSubjectLocus(), $isaDist );
	    push( @$annoISs, $annoIS );
	}
    }

    foreach my $hit3 (@$hits3){
	if( $flag3->{$hit3->getQueryID()} == 0 ){
	    my $annoIS = getAnnoIS3( $hit3->getSubjectLocus(), $isaDist );
	    push( @$annoISs, $annoIS );
	}
    }
    print STDERR "\t" . scalar(@$annoISs) . " IS positions were supported by single or paired IS-adjacent sequence(s).\n" ;

    return $annoISs;
}

sub getAnnoISPair {
    my $locus5 = shift;
    my $locus3 = shift;
    my $dist = shift;
    return 0 if $locus5->getSeqID() ne $locus3->getSeqID();
    my $status = "PAIR_SUPPORT";
    $status   .= "_OVERLAP" if $locus5->isOverlap( $locus3 );
    $status   .= "_INVALID_STRNAD" if $locus5->getStrand() ne $locus3->getStrand();
    my $minEnd   = Utils::Calc::min( [$locus5->getEnd(),   $locus3->getEnd()] );
    my $maxStart = Utils::Calc::max( [$locus5->getStart(), $locus3->getStart()] );
    my $isaDist  = $maxStart-$minEnd;
    return 0 if $isaDist > $dist;
    my $isLocus  = new Sequence::Locus();
    $isLocus->setSeqID( $locus5->getSeqID() );
    $isLocus->setStart( $minEnd );
    $isLocus->setEnd( $maxStart );
    $isLocus->setStrand( $locus5->getStrand() );
    return { locus => $isLocus, status => $status };
}

sub getAnnoIS5 {
    my $locus   = shift;
    my $dist    = shift;
    my $isLocus = new Sequence::Locus();
    if( $locus->getStrand() eq '+' ){
	$isLocus->setStart( $locus->getEnd() );
	$isLocus->setEnd( $locus->getEnd() + $dist );
    } else {
	$isLocus->setStart( $locus->getStart() - $dist );
	$isLocus->setEnd( $locus->getStart() );
    }
    $isLocus->setSeqID( $locus->getSeqID() );
    $isLocus->setStrand( $locus->getStrand() );
    return { locus => $isLocus, status => "SINGLE_SUPPORT_5" };
}

sub getAnnoIS3 {
    my $locus   = shift;
    my $dist    = shift;
    my $isLocus = new Sequence::Locus();
    if( $locus->getStrand() eq '+' ){
	$isLocus->setStart( $locus->getStart() - $dist );
	$isLocus->setEnd( $locus->getStart() );
    } else {
	$isLocus->setStart( $locus->getEnd() );
	$isLocus->setEnd( $locus->getEnd() + $dist );
    }
    $isLocus->setSeqID( $locus->getSeqID() );
    $isLocus->setStrand( $locus->getStrand() );
    return { locus => $isLocus, status => "SINGLE_SUPPORT_3" };
}

sub outputAnnotations {
    my $file = shift;
    my $annoISs = shift;
    open OUT, ">$file" or die "Can't open $file to write.";
    foreach my $annoIS (sort{$a->{locus}->getStart() <=> $b->{locus}->getStart()} @$annoISs){
	my $locus = $annoIS->{locus};
	print OUT $annoIS->{nodeID} . "\t"
	    . $annoIS->{isName} . "\t"
	    . $annoIS->{status} . "\t"
	    . $locus->{start}   . "\t"
	    . $locus->{end}     . "\n";
    }
    close OUT;
}
